/******************************************************
All parts of this code NOT already (c) by Microsoft are
(c) 2002 - 2003 by Daniel Vollmer <maven@maven.de> and
are licensed under the GNU Public License V2.
*******************************************************/

#define STRICT
#include <Windows.h>
#include <commctrl.h>
#include <commdlg.h>
#include <tchar.h>
#include <stdio.h>
#include <d3dx9.h>
#include "DXUtil.h"
#include "D3DEnumeration.h"
#include "D3DSettings.h"
#include "D3DApp.h"
#include "D3DFont.h"
#include "D3DUtil.h"
#include "resource.h"

#include "LoadWKO.h"

#define BENCH_WARMUP 100 // after how many frames we start timing
#define BENCH_FRAMES 1347 // semi-arbitrary number

#define NUM_PROCTEX 2 // 2 procedural textures (one heightmap, one normal map)

typedef struct {
	D3DXVECTOR4	Pos;
	D3DXVECTOR2	Tex;
	static const DWORD FVF;
} VERTEX;
const DWORD VERTEX::FVF = D3DFVF_XYZRHW | D3DFVF_TEX1;

typedef struct {
	D3DXVECTOR4 tex0;
	D3DXVECTOR4 tex1;
} BLOB_VERTEX;

//-----------------------------------------------------------------------------
// Name: class CMyD3DApplication
// Desc: Application class. The base class (CD3DApplication) provides the 
//       generic functionality needed in all Direct3D samples. CMyD3DApplication 
//       adds functionality specific to this sample program.
//-----------------------------------------------------------------------------
class CMyD3DApplication : public CD3DApplication
{
    CD3DFont				*m_pFont;
    CD3DArcBall				m_ArcBall;// mouse rotation utility
    D3DXVECTOR3				m_vObjectCenter;    // Center of bounding sphere of object
    float					m_fObjectRadius;    // Radius of bounding sphere of object
	DWORD					m_FrameCount;

	LPDIRECT3DCUBETEXTURE9	m_pCubeMap;

	LPDIRECT3DTEXTURE9		m_pNormalMap, m_pNormalMapFlat;

	DWORD					m_ProceduralTextureSize;
	LPDIRECT3DTEXTURE9		m_pProceduralTextures[NUM_PROCTEX]; // Old Height, New Height, Normal Map
	LPDIRECT3DSURFACE9		m_pProceduralSurfaces[NUM_PROCTEX];
	D3DVIEWPORT9			m_ProceduralViewport;
	LPD3DXRENDERTOSURFACE	m_pProceduralRTS;
	VERTEX					m_ProceduralVertices[4];
	DWORD					m_ProceduralIndex; // Idx = Old Height, (Idx + 1) % 3 = New Height, (Idx + 2) % 3 = Normal Map
	DWORD					m_NumBlobs, m_WantBlobs;
	LPDIRECT3DVERTEXBUFFER9 m_pBlobVB;
	LPDIRECT3DVERTEXDECLARATION9	m_pBlobVD;

	LPDIRECT3DTEXTURE9		m_pRTT;
    LPDIRECT3DSURFACE9      m_pRTTSurf;
	LPD3DXRENDERTOSURFACE	m_pRTS;
	D3DVIEWPORT9			m_RTViewport;

	TCHAR					m_sMeshName[512];
	
	float					m_fEta; // Index of Refraction (relative)

	float					m_fStartTime; // used for benchmark

	LPD3DXMESH				m_pMesh, m_pMeshSkyBox;
	
	LPD3DXEFFECT			m_pEffectRefraction, m_pEffectProceduralTexture, m_pEffectBlobs;

	D3DXVECTOR3				m_vRadii;
	D3DXVECTOR3				m_vObjPos, m_vCamPos; // all in world-space
	D3DXVECTOR4				m_vFilterColor;
	
	D3DXMATRIX				m_matWorld, m_matView, m_matProj, m_matAll;
    
	DWORD					m_NormalMapState; // 0 ^= dynamic, 1 ^= fixed, 2 ^= none
	bool					m_bWireframe, m_bMoveCam, m_bBenchmark;

	HRESULT LoadMesh(const char* Name);
	HRESULT UpdateWater(DWORD Src, DWORD Dst);
	HRESULT GenerateNormalMap(DWORD Src, DWORD Dst);
	void SetNumBlobs(DWORD Num);
	HRESULT DrawBlobs(DWORD Idx);
	bool WriteBenchmark(const char* Name, FLOAT start, FLOAT end);
protected:
    HRESULT OneTimeSceneInit();
    HRESULT InitDeviceObjects();
    HRESULT ConfirmDevice( D3DCAPS9* pCaps, DWORD dwBehavior, 
		D3DFORMAT adapterFormat, D3DFORMAT backBufferFormat );
    HRESULT RestoreDeviceObjects();
    HRESULT InvalidateDeviceObjects();
    HRESULT DeleteDeviceObjects();
    HRESULT Render();
    HRESULT FrameMove();
    HRESULT FinalCleanup();
    LRESULT MsgProc( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam );

public:
    CMyD3DApplication(bool Benchmark = false);
};

//-----------------------------------------------------------------------------
// Name: WinMain()
// Desc: Entry point to the program. Initializes everything, and goes into a
//       message-processing loop. Idle time is used to render the scene.
//-----------------------------------------------------------------------------
INT WINAPI WinMain( HINSTANCE hInst, HINSTANCE, LPSTR lpCmdLine, INT )
{
    CMyD3DApplication d3dApp(toupper(*lpCmdLine) == 'B');

    InitCommonControls();
    if( FAILED( d3dApp.Create( hInst ) ) )
        return 0;
    return d3dApp.Run();
}


//-----------------------------------------------------------------------------
// Name: CMyD3DApplication()
// Desc: Application constructor. Sets attributes for the app.
//-----------------------------------------------------------------------------
CMyD3DApplication::CMyD3DApplication(bool Benchmark)
{
	m_bBenchmark = Benchmark;
    m_strWindowTitle = _T("Translucency [(c) maven@maven.de]");
    m_d3dEnumeration.AppUsesDepthBuffer = FALSE;
    m_d3dEnumeration.AppRequiresFullscreen = m_bBenchmark;
    m_bShowCursorWhenFullscreen = FALSE;
	
	m_pCubeMap = NULL;

	m_fStartTime = 0.0f;

	m_fEta = 0.9f; // relative index of refraction
	m_vFilterColor = D3DXVECTOR4(0.6f, 0.6f, 0.3f, 1.0f); // we filter less blue light than red & green

	m_pMesh = m_pMeshSkyBox = NULL;
	
	m_pNormalMap = m_pRTT = m_pNormalMapFlat = NULL;
	m_pRTTSurf = NULL;
	m_pRTS = NULL;

	m_ProceduralTextureSize = 128;
	m_ProceduralIndex = 0;
	m_pProceduralRTS = NULL;
	for (DWORD i = 0; i < NUM_PROCTEX; i++)
	{
		m_pProceduralTextures[i] = NULL;
		m_pProceduralSurfaces[i] = NULL;
	}
	m_NumBlobs = 0;
	m_WantBlobs = 512;
	m_pBlobVB = NULL;
	m_pBlobVD = NULL;
	m_NormalMapState = 0;

	m_pEffectRefraction = m_pEffectProceduralTexture = m_pEffectBlobs = NULL;

    m_pFont = new CD3DFont(_T("Arial"), 12, D3DFONT_BOLD);

    _tcscpy(m_sMeshName, _T("cube.x"));
	
    m_bWireframe = false;
	m_bMoveCam = true;
}



HRESULT CMyD3DApplication::LoadMesh(const TCHAR* Name)
{
	typedef struct {
		float	pos[3];
		float	nor[3];
		float	tex[2];
		float	tang[3];
	//	float	bino[3]; // is generated in the vertex shader
	} t_vertex;
	
	HRESULT hr;
	LPD3DXBUFFER adj;
	LPD3DXMESH sysmesh;
	static const D3DVERTEXELEMENT9 decl[] =
	{
		{ 0, 0, D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_POSITION, 0 }, 
		{ 0, 12, D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_NORMAL, 0 },  
		{ 0, 24, D3DDECLTYPE_FLOAT2, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TEXCOORD, 0 },
		{ 0, 32, D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TANGENT, 0 }, 
//		{ 0, 44, D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_BINORMAL, 0 }, 
		D3DDECL_END()
	};
	
	SAFE_RELEASE(m_pMesh);

	// mesh
	if (FAILED(hr = D3DXLoadMeshFromX(Name, D3DXMESH_SYSTEMMEM, m_pd3dDevice, &adj, NULL, NULL, NULL, &sysmesh)))
		return hr;

    void *mvb;
	
	sysmesh->LockVertexBuffer(D3DLOCK_READONLY, &mvb);
	D3DXComputeBoundingSphere((D3DXVECTOR3*)mvb, sysmesh->GetNumVertices(), sysmesh->GetNumBytesPerVertex(), &m_vObjectCenter, &m_fObjectRadius);
	sysmesh->UnlockVertexBuffer();
	
	//reform the mesh to one which will have space for the tangent vectors
    if(FAILED(hr = sysmesh->CloneMesh(D3DXMESH_MANAGED, decl, m_pd3dDevice, &m_pMesh)))
	{
		SAFE_RELEASE(adj);
		SAFE_RELEASE(sysmesh);
		return hr;
	}

	SAFE_RELEASE(sysmesh);

    //compute the normals
    if(FAILED(hr = D3DXComputeNormals(m_pMesh, NULL)))
	{
		SAFE_RELEASE(adj);
		return hr;
	}

	if (m_pMesh->GetNumFaces() == 12 && m_pMesh->GetNumVertices() == 24)
	{ // hard-coded tangent vectors (as the texture mapping is not continuous across the edges of the cube)
		t_vertex *vp;
		int i;

		m_pMesh->LockVertexBuffer(0, (void**)&vp);

		for (i = 0; i < 6 * 4; i++)
		{
			vp[i].tang[0] = vp[i].tang[1] = vp[i].tang[2] = 0.0f;
//			vp[i].bino[0] = vp[i].bino[1] = vp[i].bino[2] = 0.0f;
		}

#define SET(i, n, j, v) vp[i].n[j] = vp[i+1].n[j] = vp[i+2].n[j] = vp[i+3].n[j] = v
		SET(0, tang, 2, -1.0f); //SET(0, bino, 1, -1.0f);
		SET(4, tang, 0, 1.0f); //SET(4, bino, 2, 1.0f);
		SET(8, tang, 2, 1.0f); //SET(8, bino, 1, 1.0f);
		SET(12, tang, 0, 1.0f); //SET(12, bino, 2, -1.0f);
		SET(16, tang, 1, -1.0f); //SET(16, bino, 1, 1.0f);
		SET(20, tang, 1, 1.0f); //SET(20, bino, 1, 1.0f);
#undef SET

		m_pMesh->UnlockVertexBuffer();
	}
	else if(FAILED(hr = D3DXComputeTangent(m_pMesh, 0, 0, 0, TRUE, (DWORD*)adj->GetBufferPointer())))
	{
		SAFE_RELEASE(adj);
		return hr;
	}

    m_pMesh->OptimizeInplace(D3DXMESHOPT_VERTEXCACHE | D3DXMESHOPT_COMPACT, (DWORD*)adj->GetBufferPointer(), NULL, NULL, NULL);

	adj->Release();
	return S_OK;
}

HRESULT CMyD3DApplication::UpdateWater(DWORD Src, DWORD Dst)
{
    if (SUCCEEDED(m_pProceduralRTS->BeginScene(m_pProceduralSurfaces[Dst % NUM_PROCTEX], &m_ProceduralViewport)))
    {
		UINT iPass, cPasses;
		
		m_pEffectProceduralTexture->SetTexture("tSrc", m_pProceduralTextures[Src]);
		
		m_pEffectProceduralTexture->SetTechnique("UpdateHeightMap");
		m_pEffectProceduralTexture->Begin(&cPasses, 0);
        for (iPass = 0; iPass < cPasses; iPass++)
        {
            m_pEffectProceduralTexture->Pass(iPass);
			m_pd3dDevice->SetFVF(VERTEX::FVF);
			m_pd3dDevice->DrawPrimitiveUP(D3DPT_TRIANGLESTRIP, 2, m_ProceduralVertices, sizeof(VERTEX));
        }
        m_pEffectProceduralTexture->End();
		        
        m_pProceduralRTS->EndScene(0);
	}
	return S_OK;	
}

HRESULT CMyD3DApplication::GenerateNormalMap(DWORD Src, DWORD Dst)
{
    if (SUCCEEDED(m_pProceduralRTS->BeginScene(m_pProceduralSurfaces[Dst % NUM_PROCTEX], &m_ProceduralViewport)))
    {
		UINT iPass, cPasses;
		
		m_pEffectProceduralTexture->SetTexture("tSrc", m_pProceduralTextures[Src]);
		
		m_pEffectProceduralTexture->SetTechnique("GenerateNormalMap");
		m_pEffectProceduralTexture->Begin(&cPasses, 0);
        for (iPass = 0; iPass < cPasses; iPass++)
        {
            m_pEffectProceduralTexture->Pass(iPass);
			m_pd3dDevice->SetFVF(VERTEX::FVF);
			m_pd3dDevice->DrawPrimitiveUP(D3DPT_TRIANGLESTRIP, 2, m_ProceduralVertices, sizeof(VERTEX));
        }
        m_pEffectProceduralTexture->End();
		        
        m_pProceduralRTS->EndScene(0);
	}
	return S_OK;	
}

void CMyD3DApplication::SetNumBlobs(DWORD Num)
{
	SAFE_RELEASE(m_pBlobVB);
	SAFE_RELEASE(m_pBlobVD);
	m_NumBlobs = Num;
	if (Num > 0)
	{
		static const D3DVERTEXELEMENT9 decl[] =
		{
			{ 0, 0, D3DDECLTYPE_FLOAT4, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TEXCOORD, 0 },
			{ 0, 16, D3DDECLTYPE_FLOAT4, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TEXCOORD, 1 },
			D3DDECL_END()
		};
		BLOB_VERTEX *bvs;

		if (FAILED(m_pd3dDevice->CreateVertexBuffer(Num * sizeof(BLOB_VERTEX), D3DUSAGE_WRITEONLY | D3DUSAGE_POINTS,
			0, D3DPOOL_DEFAULT, &m_pBlobVB, NULL)))
			return;
		
		m_pd3dDevice->CreateVertexDeclaration(decl, &m_pBlobVD);

		if (m_bBenchmark)
			srand(BENCH_FRAMES); // same for every run
		else
		{
			LARGE_INTEGER tt;

			tt.LowPart = (DWORD)(m_fTime * 16.0f);
			QueryPerformanceCounter(&tt);
			srand(tt.LowPart);
		}

		if (SUCCEEDED(m_pBlobVB->Lock(0, 0, (void**)&bvs, 0)))
		{
			for (DWORD i = 0; i < Num; i++)
			{
				float phase, ofs, amp, misc;

				phase = -1.4f + 2.8f * (float)rand() / ((float)RAND_MAX); // between -1.4f and 1.4f
				ofs = D3DX_PI * (float)rand() / (0.5f * (float)RAND_MAX); // between 0.0f and 2.0f * pi
				amp = 0.1f + 1.5f * (float)rand() / ((float)RAND_MAX); // between 0.1f and 1.6f
				misc = -1.0f + (float)rand() / (0.5f * (float)RAND_MAX); // betweem -1.0f and 1.0f
				bvs[i].tex0 = D3DXVECTOR4(phase, ofs, amp, misc);

				phase = -1.4f + 2.8f * (float)rand() / ((float)RAND_MAX); // between -1.4f and 1.4f
				ofs = D3DX_PI * (float)rand() / (0.5f * (float)RAND_MAX); // between 0.0f and 2.0f * pi
				amp = 0.1f + 1.5f * (float)rand() / ((float)RAND_MAX); // between 0.1f and 1.6f
				misc = -1.0f + (float)rand() / (0.5f * (float)RAND_MAX); // betweem -1.0f and 1.0f
				bvs[i].tex1 = D3DXVECTOR4(phase, ofs, amp, misc);

/*				bvs[i].tex0.z = 1.0f;
				bvs[i].tex0.w = 0.0f;
				bvs[i].tex1.z = 1.0f;
				bvs[i].tex1.w = 0.0f;
				bvs[i].tex0.x = bvs[i].tex1.x;
				bvs[i].tex0.y = bvs[i].tex1.y;*/
			}
			m_pBlobVB->Unlock();
		}
	}
}


HRESULT CMyD3DApplication::DrawBlobs(DWORD Idx)
{
    if (m_NumBlobs > 0 && SUCCEEDED(m_pProceduralRTS->BeginScene(m_pProceduralSurfaces[Idx % NUM_PROCTEX], &m_ProceduralViewport)))
    {
		FLOAT time = m_bBenchmark ? m_FrameCount / 32.0f : m_fTime;
		UINT iPass, cPasses;
		
		m_pEffectBlobs->SetFloat("fTime", time);
		
		m_pd3dDevice->SetVertexDeclaration(m_pBlobVD);
		m_pd3dDevice->SetStreamSource(0, m_pBlobVB, 0, sizeof(BLOB_VERTEX));
		m_pEffectBlobs->SetTechnique("DrawBlobs");
		m_pEffectBlobs->Begin(&cPasses, 0);
        for (iPass = 0; iPass < cPasses; iPass++)
        {
            m_pEffectBlobs->Pass(iPass);
			m_pd3dDevice->DrawPrimitive(D3DPT_POINTLIST, 0, m_NumBlobs);
        }
        m_pEffectBlobs->End();
		
		m_pProceduralRTS->EndScene(0);
	}
	return S_OK;
}


bool CMyD3DApplication::WriteBenchmark(const char* Name, FLOAT start, FLOAT end)
{
	D3DADAPTER_IDENTIFIER9 adapt;
	FILE *f;
	
	// the sample framework doesn't ask for WHQL
	if (FAILED(m_pD3D->GetAdapterIdentifier(m_d3dSettings.AdapterOrdinal(), D3DENUM_WHQL_LEVEL, &adapt)))
		return FALSE;

	f = fopen(Name, "wt");
	if (f)
	{

		fprintf(f, "Translucency 0.2 [(c) maven@maven.de]\n");
		fprintf(f, "DeviceID: %08x:%08x:%08x:%08x\n", adapt.VendorId, adapt.DeviceId, adapt.SubSysId, adapt.Revision);
		fprintf(f, "Driver: %s (%s)\n", adapt.Description, adapt.Driver);
		fprintf(f, "Driver Version: %i.%i.%i.%i\n",	HIWORD(adapt.DriverVersion.HighPart), LOWORD(adapt.DriverVersion.HighPart),
			HIWORD(adapt.DriverVersion.LowPart), LOWORD(adapt.DriverVersion.LowPart));
		fprintf(f, "WHQL Certified: %i\n", adapt.WHQLLevel > 0);
		
		fprintf(f, "%f FPS @ %ix%ix%i\n", (float)(BENCH_FRAMES - BENCH_WARMUP) / (end - start),
			m_d3dsdBackBuffer.Width, m_d3dsdBackBuffer.Height, m_d3dsdBackBuffer.Format);

		fclose(f);
		return true;
	}
	return false;
}



//-----------------------------------------------------------------------------
// Name: OneTimeSceneInit()
// Desc: Called during initial app startup, this function performs all the
//       permanent initialization.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::OneTimeSceneInit()
{
    // Set cursor to indicate that user can move the object with the mouse
#ifdef _WIN64
    SetClassLongPtr(m_hWnd, GCLP_HCURSOR, (LONG_PTR)LoadCursor(NULL, IDC_SIZEALL));
#else
    SetClassLong(m_hWnd, GCL_HCURSOR, HandleToLong(LoadCursor(NULL, IDC_SIZEALL)));
#endif
	if (m_bBenchmark)
		m_d3dSettings.SetPresentInterval(D3DPRESENT_INTERVAL_IMMEDIATE);

    return S_OK;
}




//-----------------------------------------------------------------------------
// Name: FrameMove()
// Desc: Called once per frame, the call is the entry point for animating
//       the scene.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::FrameMove()
{
	D3DXMATRIX temp, normal_scl;
	D3DXVECTOR3 wpos;

     // set up world matrix
    D3DXMatrixTranslation(&m_matWorld, -m_vObjectCenter.x, -m_vObjectCenter.y, -m_vObjectCenter.z); // adjust pivot
    D3DXMatrixMultiply(&m_matWorld, &m_matWorld, m_ArcBall.GetRotationMatrix());
    D3DXMatrixMultiply(&m_matWorld, &m_matWorld, m_ArcBall.GetTranslationMatrix());
	m_pEffectRefraction->SetMatrix("mWorld", &m_matWorld);

	wpos = m_vObjectCenter;
	D3DXVec3TransformCoord(&wpos, &wpos, &m_matWorld);
	m_vObjPos = D3DXVECTOR3(wpos.x, wpos.y, wpos.z);
	
	// set up view matrix
	if (m_bMoveCam)
	{
		FLOAT time = m_bBenchmark ? m_FrameCount / 32.0f : m_fTime;

		m_vCamPos = D3DXVECTOR3(-sinf(fmodf(time, 2.0f * D3DX_PI)), -cosf(fmodf(time * 0.7f, 2.0f * D3DX_PI)), cosf(fmodf(time * 0.6f, 2.0f * D3DX_PI)));
		m_vCamPos *= m_fObjectRadius * (2.5f + sinf(fmodf(time, 2.0f * D3DX_PI)));
	}
	else
		m_vCamPos = D3DXVECTOR3(0.0f, 0.0f, -2.0f * m_fObjectRadius);
	D3DXMatrixLookAtLH(&m_matView, &m_vCamPos, &m_vObjPos, &D3DXVECTOR3(0, 1, 0));
	m_pEffectRefraction->SetVector("vCamPos", &D3DXVECTOR4(m_vCamPos.x, m_vCamPos.y, m_vCamPos.z, 1.0f));
	
    // for the SkyBox (the tranpose is akin to computing the inverse as the matrix is orthogonal)
	m_pEffectRefraction->SetMatrixTranspose("mViewI", &m_matView);
	
	D3DXMatrixInverse(&temp, NULL, &m_matWorld);
	m_pEffectRefraction->SetMatrixTranspose("mWorldIT", &temp);
	
	D3DXMatrixMultiply(&m_matAll, &m_matWorld, &m_matView);	
	D3DXMatrixMultiply(&m_matAll, &m_matAll, &m_matProj);
	m_pEffectRefraction->SetMatrix("mWorldViewProj", &m_matAll);
	
	wpos = m_vObjPos - m_vCamPos;
	m_pEffectRefraction->SetVector("vZZZ", &D3DXVECTOR4(D3DXVec3Length(&wpos) - m_fObjectRadius, 1.0f / (2.0f * m_fObjectRadius), 2.0f * m_fObjectRadius, 1.0f));

	return S_OK;
}




//-----------------------------------------------------------------------------
// Name: Render()
// Desc: Called once per frame, the call is the entry point for 3d
//       rendering. This function sets up render states, clears the
//       viewport, and renders the scene.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::Render()
{
	if (m_bBenchmark && m_FrameCount == BENCH_WARMUP)
		m_fStartTime = m_fTime;
	
	if (m_NormalMapState == 0)
	{
		UpdateWater(m_ProceduralIndex, (m_ProceduralIndex + 1) % NUM_PROCTEX);
		GenerateNormalMap((m_ProceduralIndex + 1) % NUM_PROCTEX, (m_ProceduralIndex + 0) % NUM_PROCTEX);
		m_pEffectRefraction->SetTexture("tNormal", m_pProceduralTextures[(m_ProceduralIndex + 0) % NUM_PROCTEX]);
		DrawBlobs((m_ProceduralIndex + 1) % NUM_PROCTEX);
		m_ProceduralIndex = (m_ProceduralIndex + 1) % NUM_PROCTEX;
	}
	else if (m_NormalMapState == 1)
		m_pEffectRefraction->SetTexture("tNormal", m_pNormalMap);
	else
		m_pEffectRefraction->SetTexture("tNormal", m_pNormalMapFlat);
	
	// render the back-faces (with normals and depth) to the RenderTarget
    if (SUCCEEDED(m_pRTS->BeginScene(m_pRTTSurf, &m_RTViewport)))
    {
		UINT iPass, cPasses;

        m_pEffectRefraction->SetTechnique("StoreBackFaceNormalPass");
//		m_pd3dDevice->Clear(0L, NULL, D3DCLEAR_TARGET, 0xFFFFFFFF, 1.0f, 0L);
		m_pEffectRefraction->Begin(&cPasses, 0);
        for (iPass = 0; iPass < cPasses; iPass++)
        {
            m_pEffectRefraction->Pass(iPass);

            m_pMesh->DrawSubset(0);
        }
        m_pEffectRefraction->End();
		
		// End the scene.
        m_pRTS->EndScene(0);
    }
	
	if (SUCCEEDED(m_pd3dDevice->BeginScene()))
    {
		UINT iPass, cPasses;
		float eta0 = powf((1.0f - m_fEta), 2.0f) / powf((1.0f + m_fEta), 2.0f);
		
		m_pEffectRefraction->SetVector("vEta", &D3DXVECTOR4(m_fEta, 1.0f / m_fEta, eta0, 1.0f - eta0));
		m_pEffectRefraction->SetVector("vFilterColor", &m_vFilterColor);
		m_pEffectRefraction->SetTexture("tRenderTarget", m_pRTT);

		if (m_bWireframe)
			m_pd3dDevice->Clear(0L, NULL, D3DCLEAR_TARGET, 0x00000000, 1.0f, 0L);
        
		m_pEffectRefraction->SetTechnique("SkyBoxPass");
		m_pEffectRefraction->Begin(&cPasses, 0);
        for (iPass = 0; iPass < cPasses; iPass++)
        {
            m_pEffectRefraction->Pass(iPass);

            m_pMeshSkyBox->DrawSubset(0);
        }
        m_pEffectRefraction->End();
		
		m_pEffectRefraction->SetTechnique("FinalPass");
		m_pEffectRefraction->Begin(&cPasses, 0);
        for (iPass = 0; iPass < cPasses; iPass++)
        {
            m_pEffectRefraction->Pass(iPass);

            m_pMesh->DrawSubset(0);
        }
        m_pEffectRefraction->End();
        
		if (!m_bBenchmark)
		{ // Output statistics
			TCHAR buf[256];

			m_pFont->DrawText(2, 0, D3DCOLOR_XRGB(255,255,0), m_strFrameStats);
			m_pFont->DrawText(2, 20, D3DCOLOR_XRGB(255,255,0), m_strDeviceStats);
			if (m_NormalMapState == 0)
				_stprintf(buf, "Eta: %.02f Particles: %i", m_fEta, m_NumBlobs);
			else
				_stprintf(buf, "Eta: %.02f", m_fEta);
			m_pFont->DrawText(2, 40, D3DCOLOR_XRGB(255,255,0), buf);
		}
		m_pd3dDevice->EndScene();
		if (m_bBenchmark && ++m_FrameCount == BENCH_FRAMES)
		{
			WriteBenchmark("bench.txt", m_fStartTime, m_fTime);
			PostQuitMessage(0);
		}
	}	
    return S_OK;
}

//-----------------------------------------------------------------------------
// Name: InitDeviceObjects()
// Desc: This creates all device-dependent managed objects, such as managed
//       textures and managed vertex buffers.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::InitDeviceObjects()
{
	HRESULT hr;

    // Initialize the font's internal textures
    m_pFont->InitDeviceObjects(m_pd3dDevice);

	m_fObjectRadius = sqrtf(0.75f);
	m_vObjPos = D3DXVECTOR3(0.0f, 0.0f, 0.0f);

    m_pCubeMap = LoadCubeMapWKO("envmap_stpeters.wko", LTX_CREATE_MIPMAPS, m_pd3dDevice, D3DFMT_UNKNOWN);

	m_pNormalMap = LoadTextureMapWKO("normal_tree.wko", LTX_CREATE_MIPMAPS, m_pd3dDevice, D3DFMT_UNKNOWN);
	
	m_pNormalMapFlat = LoadTextureMapWKO("normal_flat.wko", LTX_CREATE_MIPMAPS, m_pd3dDevice, D3DFMT_UNKNOWN);

	// skybox
	if (FAILED(hr = D3DXLoadMeshFromX("skybox.x", D3DXMESH_MANAGED, m_pd3dDevice, NULL, NULL, NULL, NULL, &m_pMeshSkyBox)))
		return hr;

	LoadMesh(m_sMeshName);
	
	return S_OK;
}

//-----------------------------------------------------------------------------
// Name: RestoreDeviceObjects()
// Desc: Restore device-memory objects and state after a device is created or
//       resized.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::RestoreDeviceObjects()
{
    LPD3DXBUFFER pBufferErrors = NULL;
    HRESULT hr;

	m_FrameCount = 0;
    m_ArcBall.SetWindow(m_d3dsdBackBuffer.Width, m_d3dsdBackBuffer.Height, 0.85f);
    m_ArcBall.SetRadius(m_fObjectRadius);
    
	// Restore the font
    if (FAILED(hr = m_pFont->RestoreDeviceObjects()))
		return hr;
	
	m_pd3dDevice->SetRenderState(D3DRS_LIGHTING, FALSE);
	m_pd3dDevice->SetRenderState(D3DRS_AMBIENT, 0xffffffff);
    m_pd3dDevice->SetRenderState(D3DRS_ZENABLE, FALSE);
    m_pd3dDevice->SetRenderState(D3DRS_DITHERENABLE, TRUE);
    if (m_bWireframe)
        m_pd3dDevice->SetRenderState(D3DRS_FILLMODE, D3DFILL_WIREFRAME);
    else
        m_pd3dDevice->SetRenderState(D3DRS_FILLMODE, D3DFILL_SOLID);    
    
    // Setup the projection matrix
    float fAspect = (float)m_d3dsdBackBuffer.Width / (float)m_d3dsdBackBuffer.Height;
    D3DXMatrixPerspectiveFovLH(&m_matProj, D3DX_PI / 4.0f, fAspect, m_fObjectRadius / 64.0f, m_fObjectRadius * 200.0f);
    m_pd3dDevice->SetTransform(D3DTS_PROJECTION, &m_matProj);
    
	// create RT
	// we are calling CreateTexture directly so we can get a non-pow2 texture
	if (FAILED(hr = m_pd3dDevice->CreateTexture(m_d3dsdBackBuffer.Width, m_d3dsdBackBuffer.Height, 1,
		D3DUSAGE_RENDERTARGET, D3DFMT_A8R8G8B8, D3DPOOL_DEFAULT, &m_pRTT, NULL)))
		return hr;
    m_pRTT->GetSurfaceLevel(0, &m_pRTTSurf);
	
	D3DSURFACE_DESC desc;
    m_pRTTSurf->GetDesc(&desc);

    if (FAILED(hr = D3DXCreateRenderToSurface(m_pd3dDevice, desc.Width, desc.Height,
        desc.Format, FALSE, D3DFMT_UNKNOWN, &m_pRTS)))
		return hr;

	m_pd3dDevice->GetViewport(&m_RTViewport);
	m_RTViewport.X = m_RTViewport.Y = 0;

	if (FAILED(hr = D3DXCreateEffectFromFile(m_pd3dDevice, "translucent.h", NULL, NULL, 0, NULL, &m_pEffectRefraction, &pBufferErrors)))
	{
		OutputDebugString("--- translucent.h ---\n");
		if (pBufferErrors)
		{
			OutputDebugString((TCHAR*)pBufferErrors->GetBufferPointer());
			pBufferErrors->Release();
		}
		return hr;
	}
	m_pEffectRefraction->SetVector("vTexCoordBias", &D3DXVECTOR4(0.5f + 0.5f / (float)desc.Width, 0.5f + 0.5f / (float)desc.Height, 0.0f, 0.0f));
	m_pEffectRefraction->SetTexture("tNormal", m_pNormalMap);
	m_pEffectRefraction->SetTexture("tCube", m_pCubeMap);
	
	
	
	// create stuff for the procedural textures
	for (DWORD i = 0; i < NUM_PROCTEX; i++)
	{
		if (FAILED(D3DXCreateTexture(m_pd3dDevice, m_ProceduralTextureSize, m_ProceduralTextureSize, 1, D3DUSAGE_RENDERTARGET,
			D3DFMT_A8R8G8B8, D3DPOOL_DEFAULT, &m_pProceduralTextures[i])))
			return hr;

		m_pProceduralTextures[i]->GetSurfaceLevel(0, &m_pProceduralSurfaces[i]);
	}
	m_pProceduralSurfaces[0]->GetDesc(&desc);

    if (FAILED(hr = D3DXCreateRenderToSurface(m_pd3dDevice, desc.Width, desc.Height,
        desc.Format, FALSE, D3DFMT_UNKNOWN, &m_pProceduralRTS)))
		return hr;

	m_ProceduralViewport.X = 0;
	m_ProceduralViewport.Y = 0;
	m_ProceduralViewport.Width = desc.Width;
	m_ProceduralViewport.Height = desc.Height;
	m_ProceduralViewport.MinZ = 0.0f;
	m_ProceduralViewport.MaxZ = 1.0f;

	// clear all textures
	for (i = 0; i < NUM_PROCTEX; i++)
	{
		if (SUCCEEDED(m_pProceduralRTS->BeginScene(m_pProceduralSurfaces[i], &m_ProceduralViewport)))
		{
			m_pd3dDevice->Clear(0L, NULL, D3DCLEAR_TARGET, D3DCOLOR_ARGB(0, 128, 0, 0), 1.0f, 0L);
			m_pProceduralRTS->EndScene(0);
		}
	}

	m_ProceduralVertices[0].Pos = D3DXVECTOR4(desc.Width - 0.5f, -0.5f, 0.0f, 1.0f);
	m_ProceduralVertices[0].Tex = D3DXVECTOR2(1.0f, 0.0f);
	m_ProceduralVertices[1].Pos = D3DXVECTOR4(desc.Width - 0.5f, desc.Height - 0.5f, 0.0f, 1.0f);
	m_ProceduralVertices[1].Tex = D3DXVECTOR2(1.0f, 1.0f);
	m_ProceduralVertices[2].Pos = D3DXVECTOR4(-0.5f, -0.5f, 0.0f, 1.0f);
	m_ProceduralVertices[2].Tex = D3DXVECTOR2(0.0f, 0.0f);
	m_ProceduralVertices[3].Pos = D3DXVECTOR4(-0.5f, desc.Height - 0.5f, 0.0f, 1.0f);
	m_ProceduralVertices[3].Tex = D3DXVECTOR2(0.0f, 1.0f);
	SetNumBlobs(m_WantBlobs);
	
	if (FAILED(hr = D3DXCreateEffectFromFile(m_pd3dDevice, "procedural.h", NULL, NULL, 0, NULL, &m_pEffectProceduralTexture, &pBufferErrors)))
	{
		OutputDebugString("--- procedural.h ---\n");
		if (pBufferErrors)
		{
			OutputDebugString((TCHAR*)pBufferErrors->GetBufferPointer());
			pBufferErrors->Release();
		}
		return hr;
	}
	
	D3DXVECTOR4 ofs[2];
	ofs[0] = D3DXVECTOR4(-1.0f / (float)desc.Width, 0.0f, 1.0f / (float)desc.Width, 0.0f); // sample west & east
	ofs[1] = D3DXVECTOR4(0.0f, -1.0f / (float)desc.Height, 0.0f, 1.0f / (float)desc.Height); // sample north & south
	m_pEffectProceduralTexture->SetVectorArray("vTexOffsets", ofs, 2);
	ofs[0] = D3DXVECTOR4(-1.0f / (float)desc.Width, -1.0f / (float)desc.Height, 1.0f / (float)desc.Width, -1.0f / (float)desc.Height); // sample west & east
	ofs[1] = D3DXVECTOR4(-1.0f / (float)desc.Width, 1.0f / (float)desc.Height, 1.0f / (float)desc.Width, 1.0f / (float)desc.Height); // sample west & east
	m_pEffectProceduralTexture->SetVectorArray("vTexOffsetsDiag", ofs, 2);
	
	
	// blobs
	if (FAILED(hr = D3DXCreateEffectFromFile(m_pd3dDevice, "blobs.h", NULL, NULL, 0, NULL, &m_pEffectBlobs, &pBufferErrors)))
	{
		OutputDebugString("--- blobs.h ---\n");
		if (pBufferErrors)
		{
			OutputDebugString((TCHAR*)pBufferErrors->GetBufferPointer());
			pBufferErrors->Release();
		}
		return hr;
	}
	
    return S_OK;
}




//-----------------------------------------------------------------------------
// Name: InvalidateDeviceObjects()
// Desc: Called when the device-dependent objects are about to be lost.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::InvalidateDeviceObjects()
{
    m_pFont->InvalidateDeviceObjects();
	
	SetNumBlobs(0);
	
	SAFE_RELEASE(m_pEffectBlobs);
	SAFE_RELEASE(m_pEffectRefraction);
	SAFE_RELEASE(m_pEffectProceduralTexture);

	SAFE_RELEASE(m_pProceduralRTS);
	for (DWORD i = 0; i < NUM_PROCTEX; i++)
	{
		SAFE_RELEASE(m_pProceduralSurfaces[i]);
		SAFE_RELEASE(m_pProceduralTextures[i]);
	}
    
	SAFE_RELEASE(m_pRTS);	
	SAFE_RELEASE(m_pRTTSurf);
	SAFE_RELEASE(m_pRTT);

    return S_OK;
}




//-----------------------------------------------------------------------------
// Name: DeleteDeviceObjects()
// Desc: Called when the app is exiting, or the device is being changed,
//       this function deletes any device dependent objects.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::DeleteDeviceObjects()
{
    m_pFont->DeleteDeviceObjects();
    
	SAFE_RELEASE(m_pCubeMap);
	SAFE_RELEASE(m_pNormalMap);
	SAFE_RELEASE(m_pNormalMapFlat);

	SAFE_RELEASE(m_pMeshSkyBox);
	SAFE_RELEASE(m_pMesh);

    return S_OK;
}





//-----------------------------------------------------------------------------
// Name: FinalCleanup()
// Desc: Called during initial app startup, this function performs all the
//       permanent initialization.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::FinalCleanup()
{
    SAFE_DELETE(m_pFont);

    return S_OK;
}




//-----------------------------------------------------------------------------
// Name: MsgProc()
// Desc: Message proc function to handle key and menu input
//-----------------------------------------------------------------------------
LRESULT CMyD3DApplication::MsgProc(HWND hWnd, UINT uMsg, WPARAM wParam,
								   LPARAM lParam)
{
	if (!m_bBenchmark)
	{
		// Pass mouse messages to the ArcBall so it can build internal matrices
		m_ArcBall.HandleMouseMessages(hWnd, uMsg, wParam, lParam);

		// Handle key presses
		if (uMsg == WM_KEYDOWN)
		{
			if (wParam == (int)'W')
			{
				m_bWireframe = !m_bWireframe;
				m_pd3dDevice->SetRenderState(D3DRS_FILLMODE, m_bWireframe ? D3DFILL_WIREFRAME : D3DFILL_SOLID);
			}
			else if (wParam == (int)'C')
				m_bMoveCam = !m_bMoveCam;
			else if (wParam == (int)'D')
				m_NormalMapState = (m_NormalMapState + 1) % 3;
			else if (wParam == (int)'F')
			{ // Filter colour
				CHOOSECOLOR cc;                 // common dialog box structure 
				static COLORREF acrCustClr[16]; // array of custom colors 

				// Initialize CHOOSECOLOR 
				ZeroMemory(&cc, sizeof cc);
				cc.lStructSize = sizeof cc;
				cc.hwndOwner = m_hWnd;
				cc.lpCustColors = (LPDWORD)acrCustClr;
				cc.rgbResult = D3DCOLOR_COLORVALUE(1.0f - m_vFilterColor.z, 1.0f - m_vFilterColor.y, 1.0f - m_vFilterColor.x, 0.0f);
				cc.Flags = CC_FULLOPEN | CC_RGBINIT;
 
				if (ChooseColor(&cc) == TRUE)
				{
					m_vFilterColor.x = 1.0f - GetRValue(cc.rgbResult) / 255.0f;
					m_vFilterColor.y = 1.0f - GetGValue(cc.rgbResult) / 255.0f;
					m_vFilterColor.z = 1.0f - GetBValue(cc.rgbResult) / 255.0f;
				}
			}
 			else if (wParam == VK_ADD)
				m_fEta += 0.01f;
 			else if (wParam == VK_SUBTRACT)
				m_fEta = max(0.0f, m_fEta - 0.01f);
 			else if (wParam == VK_DIVIDE)
			{
				if (m_NumBlobs > 0)
					SetNumBlobs(m_NumBlobs / 2);
				m_WantBlobs = m_NumBlobs;
			}
 			else if (wParam == VK_MULTIPLY)
			{
				if (m_NumBlobs == 0)
					SetNumBlobs(1);
				else if (m_NumBlobs * 2 <= m_d3dCaps.MaxPrimitiveCount)
					SetNumBlobs(m_NumBlobs * 2);
				m_WantBlobs = m_NumBlobs;
			}
			else if (wParam == VK_NUMPAD0)
				m_fEta = 0.0f;
			else if (wParam == VK_NUMPAD1)
				m_fEta = 0.16f;
			else if (wParam == VK_NUMPAD2)
				m_fEta = 0.40f;
			else if (wParam == VK_NUMPAD3)
				m_fEta = 0.90f;
			else if (wParam == VK_NUMPAD4)
				m_fEta = 1.0f;
			else if (wParam == VK_NUMPAD5)
				m_fEta = 1.2f;
		}
		else if (uMsg == WM_COMMAND)
		{
            // Handle the open file command
			if (LOWORD(wParam) == IDM_OPENFILE)
			{
				OPENFILENAME ofn;
				TCHAR fname[512];

				ZeroMemory(&ofn, sizeof ofn);
				_tcscpy(fname, m_sMeshName);
				ofn.lStructSize = sizeof ofn;
				ofn.hwndOwner = m_hWnd;
				ofn.lpstrFilter = ".X Files (.x)\0*.x\0";
				ofn.lpstrFile = fname;
				ofn.nMaxFile = 512;
				ofn.Flags = OFN_FILEMUSTEXIST | OFN_PATHMUSTEXIST | OFN_NOCHANGEDIR;
				ofn.lpstrTitle = _T("Open Mesh File");
				if (GetOpenFileName(&ofn))
				{
					strcpy(m_sMeshName, fname);
					LoadMesh(ofn.lpstrFile);
				}
			}
		}
    }

    // Pass remaining messages to default handler
    return CD3DApplication::MsgProc( hWnd, uMsg, wParam, lParam );
}

//-----------------------------------------------------------------------------
// Name: ConfirmDevice()
// Desc: Called during device initialization, this code checks the device
//       for some minimum set of capabilities
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::ConfirmDevice(D3DCAPS9* pCaps, DWORD dwBehavior, 
										 D3DFORMAT adapterFormat, D3DFORMAT backBufferFormat)
{
    if ((pCaps->TextureCaps & D3DPTEXTURECAPS_CUBEMAP) != 0 &&
		((pCaps->TextureCaps & D3DPTEXTURECAPS_NONPOW2CONDITIONAL) != 0 || (pCaps->TextureCaps & D3DPTEXTURECAPS_POW2) == 0) &&
        pCaps->VertexShaderVersion >= D3DVS_VERSION(2, 0) &&
		pCaps->PixelShaderVersion >= D3DPS_VERSION(2, 0) &&
		pCaps->MaxSimultaneousTextures >= 7 /* we use texcoord 0-6 */ &&
		SUCCEEDED(m_pD3D->CheckDeviceFormat(pCaps->AdapterOrdinal, pCaps->DeviceType, adapterFormat, D3DUSAGE_RENDERTARGET, D3DRTYPE_TEXTURE, D3DFMT_A8R8G8B8))
		)
	{
		return S_OK;
    }        
	return E_FAIL;
}