/******************************************************
(c) 2002 - 2003 by Daniel Vollmer <maven@maven.de>.
Licensed under the GNU Public License V2.

Interface for my Wavelet-Image codec, which can
be found at http://www.maven.de/ -> code -> wavelet.
*******************************************************/

#include <d3dx9.h>
#include "../wavelet/wavelet.h"
#include "../wavelet/mchannel.h"
#include "LoadWKO.h"

static DWORD *LoadWKO2MemImage(char* Filename, LTX_FLAGS Flags, int *Width, int*Height)
{
	DWORD *memimage = NULL;
	int width = 0, height = 0;
	t_bit_file* bf = bit_open(Filename, "rb", 0);

	if (bf)
	{
		t_wv_dchannels* dc;
		bool ycbcr;

		ycbcr = bit_read_single(bf) != 0;
		dc = wv_init_decode_channels(0, bf);
		bit_close(bf, NULL);
		bf = NULL;

		if (dc)
		{
			int y, x, i;

			memimage = (DWORD*)malloc(dc->hdr.owidth * dc->hdr.oheight * sizeof *memimage);
			if (dc->hdr.num_channels >= 3)
			{
				wv_pel *temp;

				temp = dc->hdr.num_channels == 3 && (Flags & LTX_REPLICATE_LUMINANCE_TO_ALPHA) != 0 ? (wv_pel*)malloc(3 * dc->hdr.owidth * sizeof *temp) : NULL;

				for (y = 0; y < dc->hdr.oheight; y++)
				{
					wv_pel *src_r, *src_g, *src_b, *src_a = NULL;

					src_r = dc->channels[0] + y * dc->hdr.width;
					src_g = dc->channels[1] + y * dc->hdr.width;
					src_b = dc->channels[2] + y * dc->hdr.width;

					if (ycbcr)
					{
						if (dc->hdr.num_channels == 3 && (Flags & LTX_REPLICATE_LUMINANCE_TO_ALPHA) != 0)
						{
							memcpy(temp, dc->channels[0] + y * dc->hdr.width, dc->hdr.owidth * sizeof *temp);
							src_a = temp;
						}
						else if (dc->hdr.num_channels > 3)
							src_a = dc->channels[3] + y * dc->hdr.width;
						
						if (src_a)
						{ // clamp to [0, 255] (ycrcr->rgb conversion clamped already)
							for (x = 0; x < dc->hdr.owidth; x++)
								src_a[x] = (wv_pel)min(255, max(0, src_a[x]));
						}

						wv_ycbcr_to_rgb(dc->hdr.owidth, dc->channels[0] + y * dc->hdr.width, dc->channels[1] + y * dc->hdr.width,
							dc->channels[2] + y * dc->hdr.width);						
					}
					else 
					{ 
						for (x = 0; x < dc->hdr.owidth; x++)
						{ // clamp to [0, 255] (because we might've been using lossy compression)
							src_r[x] = (wv_pel)min(255, max(0, src_r[x]));
							src_g[x] = (wv_pel)min(255, max(0, src_g[x]));
							src_b[x] = (wv_pel)min(255, max(0, src_b[x]));
						}
						
						if ((Flags & LTX_REPLICATE_LUMINANCE_TO_ALPHA) != 0)
						{ // so we need the luminance info
							memcpy(temp, dc->channels[0] + y * dc->hdr.width, dc->hdr.owidth * sizeof *temp);
							memcpy(temp + dc->hdr.owidth, dc->channels[1] + y * dc->hdr.width, dc->hdr.owidth * sizeof *temp);
							memcpy(temp + 2 * dc->hdr.owidth, dc->channels[2] + y * dc->hdr.width, dc->hdr.owidth * sizeof *temp);
							wv_rgb_to_ycbcr(dc->hdr.owidth, temp, temp + dc->hdr.owidth, temp + 2 * dc->hdr.owidth);
							src_a = temp;
							for (x = 0; x < dc->hdr.owidth; x++)
								src_a[x] = (wv_pel)min(255, max(0, src_a[x]));
						}
					}
					
					if (src_a)
					{
						for (x = 0; x < dc->hdr.owidth; x++)
							memimage[y * dc->hdr.owidth + x] = D3DCOLOR_ARGB(src_a[x], src_r[x], src_g[x], src_b[x]);
					}
					else
					{
						for (x = 0; x < dc->hdr.owidth; x++)
							memimage[y * dc->hdr.owidth + x] = D3DCOLOR_XRGB(src_r[x], src_g[x], src_b[x]);
					}
				}
				
				if (temp != NULL)
					free(temp);
			}
			else
			{ // average the channels and replicate to all channels
				for (y = 0; y < dc->hdr.oheight; y++)
				{
					DWORD aa = 0;

					if ((Flags & LTX_REPLICATE_LUMINANCE_TO_ALPHA) != 0)
					{
						for (x = 0; x < dc->hdr.owidth; x++)
						{

							for (i = 0; i < dc->hdr.num_channels; i++)
								aa += dc->channels[i][y * dc->hdr.width + x];
							aa = min(255, max(0, aa / dc->hdr.num_channels));
							memimage[y * dc->hdr.owidth + x] = D3DCOLOR_ARGB(aa, aa, aa, aa);
						}
					}
					else
					{
						for (x = 0; x < dc->hdr.owidth; x++)
						{
							for (i = 0; i < dc->hdr.num_channels; i++)
								aa += dc->channels[i][y * dc->hdr.width + x];
							aa = min(255, max(0, aa / dc->hdr.num_channels));
							memimage[y * dc->hdr.owidth + x] = D3DCOLOR_XRGB(aa, aa, aa);
						}
					}
				}
			}
			width = dc->hdr.owidth;
			height = dc->hdr.oheight;
			wv_done_decode_channels(dc);
		}
	}
	if (memimage)
	{
		*Width = width;
		*Height = height;
	}
	else
	{
		*Width = 0;
		*Height = 0;
	}
	return memimage;
}


static HRESULT CopySYSMEMSurface2Texture(LPDIRECT3DSURFACE9 Dst, LPDIRECT3DSURFACE9 Src, LPDIRECT3DDEVICE9 Device)
{
	HRESULT hr;
	D3DSURFACE_DESC dstd, srcd;
	
	if (FAILED(hr = Dst->GetDesc(&dstd)))
		return hr;
	
	if (FAILED(hr = Src->GetDesc(&srcd)))
		return hr;
	
	if (dstd.Pool == D3DPOOL_DEFAULT)
	{ // use UpdateSurface
		if (FAILED(hr = Device->UpdateSurface(Src, NULL, Dst, NULL)))
			return hr;
	}
	else
	{ // use
		int y, ymax;
		D3DLOCKED_RECT sr, dr;

		if (FAILED(hr = Src->LockRect(&sr, NULL, D3DLOCK_READONLY)))
			return hr;

		if (FAILED(hr = Dst->LockRect(&dr, NULL, 0)))
		{
			Src->UnlockRect();
			return hr;
		}

		ymax = min(srcd.Height, dstd.Height);
		if (dstd.Format == D3DFMT_DXT1 || dstd.Format == D3DFMT_DXT2 || dstd.Format == D3DFMT_DXT3 || dstd.Format == D3DFMT_DXT4 || dstd.Format == D3DFMT_DXT5)
			ymax /= 4;
		for (y = 0; y < ymax; y++)
			memcpy((byte *)dr.pBits + y * dr.Pitch, (byte *)sr.pBits + y * sr.Pitch, min(sr.Pitch, dr.Pitch));

		if (FAILED(hr = Dst->UnlockRect()))
		{
			Src->UnlockRect();
			return hr;
		}

		if (FAILED(hr = Src->UnlockRect()))
			return hr;
	}
	return S_OK;
}


LPDIRECT3DTEXTURE9 LoadTextureMapWKO(char* Filename, LTX_FLAGS Flags, LPDIRECT3DDEVICE9 Device, D3DFORMAT Format)
{
	LPDIRECT3DTEXTURE9 tex = NULL;
	DWORD *memimage;
	int width, height;
	
	memimage = LoadWKO2MemImage(Filename, Flags, &width, &height);
	if (memimage && width > 0 && height > 0)
	{
		LPDIRECT3DTEXTURE9 systex = NULL;
		LPDIRECT3DSURFACE9 syssurf = NULL;
		int i;
		RECT dim;

		if (FAILED(D3DXCreateTexture(Device, width, height, (Flags & LTX_CREATE_MIPMAPS) == 0,
			0, Format, D3DPOOL_SYSTEMMEM, &systex)))
		{
			free(memimage);
			return NULL;
		}
			
		dim.left = dim.top = 0;
		dim.right = width;
		dim.bottom = height;
		
		// now get memimage into the surface
		if (FAILED(systex->GetSurfaceLevel(0, &syssurf)))
		{
			systex->Release();
			return NULL;
		}

		if (FAILED(D3DXLoadSurfaceFromMemory(syssurf, NULL, NULL, memimage, D3DFMT_A8R8G8B8,
			dim.right * sizeof *memimage, NULL, &dim, D3DX_DEFAULT, 0)))
		{
			syssurf->Release();
			systex->Release();
			return NULL;
		}

		syssurf->Release();
		
		if ((Flags & LTX_CREATE_MIPMAPS) != 0 && FAILED(D3DXFilterTexture(systex, NULL, 0, D3DX_DEFAULT)))
		{
			systex->Release();
			return NULL;
		}

		if (FAILED(D3DXCreateTexture(Device, dim.right, dim.bottom, (Flags & LTX_CREATE_MIPMAPS) == 0,
			((Flags & LTX_USAGE_RENDERTARGET) != 0) ? D3DUSAGE_RENDERTARGET : 0, Format,
			((Flags & (LTX_POOL_DEFAULT | LTX_USAGE_RENDERTARGET)) != 0) ? D3DPOOL_DEFAULT : D3DPOOL_MANAGED, &tex)))
		{
			systex->Release();
			return NULL;
		}

		for (i = 0; i < (int)tex->GetLevelCount(); i++)
		{
			LPDIRECT3DSURFACE9 src, dst;

			if (FAILED(systex->GetSurfaceLevel(i, &src)))
			{
				tex->Release();
				systex->Release();
				return NULL;
			}

			if (FAILED(tex->GetSurfaceLevel(i, &dst)))
			{
				src->Release();
				tex->Release();
				systex->Release();
				return NULL;
			}

			if (FAILED(CopySYSMEMSurface2Texture(dst, src, Device)))
			{
				dst->Release();
				src->Release();
				tex->Release();
				systex->Release();
				return NULL;
			}

			dst->Release();
			src->Release();
		}
		systex->Release();
	}
	if (memimage)
		free(memimage);
	return tex;
}

static void RotateMemRectCW(DWORD *Mem, int Width, int AllWidth)
{
	DWORD *tdst = (DWORD*)malloc(Width * Width * sizeof *tdst);
	int y, x;

	for (y = 0; y < Width; y++)
		for (x = 0; x < Width; x++)
			tdst[x * Width + (Width - 1 - y)] = Mem[y * AllWidth + x];
	for (y = 0; y < Width; y++)
		memcpy(Mem + y * AllWidth, tdst + y * Width, Width * sizeof *Mem);
	free(tdst);
}

LPDIRECT3DCUBETEXTURE9 LoadCubeMapWKO(char* Filename, LTX_FLAGS Flags, LPDIRECT3DDEVICE9 Device, D3DFORMAT Format)
{
	LPDIRECT3DCUBETEXTURE9 tex = NULL;
	DWORD *memimage;
	int width, height;
	
	memimage = LoadWKO2MemImage(Filename, Flags, &width, &height);

	/*
		CubeMap Layout:
		[[ 4  0  2']  (2 is rotated 90 CCW in image, so needs to be rotated 90 CW to "undo")
		 [ 3  5  1 ]]
	*/


	if (memimage && width > 0 && height > 0 && (width / 3) == (height / 2))
	{
		LPDIRECT3DCUBETEXTURE9 systex = NULL;
		LPDIRECT3DSURFACE9 syssurf = NULL;
		int i, j;
		RECT dim;

		width /= 3;
		height /= 2;
		
		if (FAILED(D3DXCreateCubeTexture(Device, width, (Flags & LTX_CREATE_MIPMAPS) == 0,
			0, Format, D3DPOOL_SYSTEMMEM, &systex)))
		{
			free(memimage);
			return NULL;
		}
			
		dim.left = dim.top = 0;
		dim.right = width;
		dim.bottom = height;
		
		for (j = 0; j < 6; j++)
		{
			static const unsigned char cube_idx[6] = {1, 5, 2 + 8, 3, 0, 4};
			int idx = cube_idx[j] & 7;
			DWORD *memsrc = memimage + (idx / 3) * 3 * width * height + (idx % 3) * width;

			// now get memimage into the surface
			if (FAILED(systex->GetCubeMapSurface((D3DCUBEMAP_FACES)j, 0, &syssurf)))
			{
				systex->Release();
				return NULL;
			}

			if ((cube_idx[j] & 8) != 0)
				RotateMemRectCW(memsrc, width, 3 * width);

			if (FAILED(D3DXLoadSurfaceFromMemory(syssurf, NULL, NULL, memsrc, D3DFMT_A8R8G8B8,
				3 * width * sizeof *memimage, NULL, &dim, D3DX_DEFAULT, 0)))
			{
				syssurf->Release();
				systex->Release();
				return NULL;
			}
			syssurf->Release();			
		}

		if ((Flags & LTX_CREATE_MIPMAPS) != 0 && FAILED(D3DXFilterTexture(systex, NULL, 0, D3DX_DEFAULT)))
		{
			systex->Release();
			return NULL;
		}

		if (FAILED(D3DXCreateCubeTexture(Device, width, (Flags & LTX_CREATE_MIPMAPS) == 0,
			((Flags & LTX_USAGE_RENDERTARGET) != 0) ? D3DUSAGE_RENDERTARGET : 0, Format,
			((Flags & (LTX_POOL_DEFAULT | LTX_USAGE_RENDERTARGET)) != 0) ? D3DPOOL_DEFAULT : D3DPOOL_MANAGED, &tex)))
		{
			systex->Release();
			return NULL;
		}

		for (i = 0; i < (int)tex->GetLevelCount(); i++)
		{
			for (j = 0; j < 6; j++)
			{
				LPDIRECT3DSURFACE9 src, dst;

				if (FAILED(systex->GetCubeMapSurface((D3DCUBEMAP_FACES)j, i, &src)))
				{
					tex->Release();
					systex->Release();
					return NULL;
				}

				if (FAILED(tex->GetCubeMapSurface((D3DCUBEMAP_FACES)j, i, &dst)))
				{
					src->Release();
					tex->Release();
					systex->Release();
					return NULL;
				}

				if (FAILED(CopySYSMEMSurface2Texture(dst, src, Device)))
				{
					dst->Release();
					src->Release();
					tex->Release();
					systex->Release();
					return NULL;
				}

				dst->Release();
				src->Release();
			}
		}

		systex->Release();
	}
	if (memimage)
		free(memimage);
	return tex;
}
