/******************************************************
(c) 2002 - 2003 by Daniel Vollmer <maven@maven.de>.
Licensed under the GNU Public License V2.
*******************************************************/

texture tSrc;

sampler sSrc = sampler_state
{
	Texture = <tSrc>;
	MipFilter = NONE;
	MinFilter = POINT;
	MagFilter = POINT;
	AddressU = Wrap;
	AddressV = Wrap;
};

float unpack_both(float4 color, out float velocity)
{
	velocity = color.b + color.a * (1 / 256);
	return color.r + color.g * (1 / 256);
/*	velocity = color.b;
	return color.r;*/
}

float unpack(float4 color)
{
	return color.r + color.g * (1 / 256);
//	return color.r;
}

float4 pack(float height, float velocity)
{
	return float4(height, frac(height * 256), velocity, frac(velocity * 256)); // keep a bit more fractional as well
//	return float4(height, 0, velocity, 0);
}

static const float c1 = 0.4;
static const float c2 = 0.48;
static const float k = 1.0;
static const float k2 = 0.15;
static const float d1 = 0.9875;

float4 vTexOffsets[2];

float4 ps_update_heightmap(float2 UV: TEXCOORD0) : COLOR
{
	float H1 = unpack(tex2D(sSrc, UV + vTexOffsets[0].xy));
	float H2 = unpack(tex2D(sSrc, UV + vTexOffsets[0].zw));
	float H3 = unpack(tex2D(sSrc, UV + vTexOffsets[1].xy));
	float H4 = unpack(tex2D(sSrc, UV + vTexOffsets[1].zw));
	float V;
	float H0 = unpack_both(tex2D(sSrc, UV), V);

	float F = k * (H1 + H2 + H3 + H4 - 4*H0) + k2 * (0.5 - H0);
	V += c1 * F;
	H0 = H0 * d1 + c2 * V;

	H0 = (3 * H0 + H1 + H2 + H3 + H4) / 7; // and smooth the height a bit

	return pack(H0, V);
}

technique UpdateHeightMap
{
	pass P0
	{
		PixelShader = compile ps_2_0 ps_update_heightmap();
	}
}



float4 vTexOffsetsDiag[2];

float4 ps_generate_normalmap(float2 UV: TEXCOORD0, uniform bool cheap) : COLOR
{ // from ShaderX^2
	float l = unpack(tex2D(sSrc, UV + vTexOffsets[0].xy));
	float r = unpack(tex2D(sSrc, UV + vTexOffsets[0].zw));
	float t = unpack(tex2D(sSrc, UV + vTexOffsets[1].xy));
	float b = unpack(tex2D(sSrc, UV + vTexOffsets[1].zw));

	float dX, dY;
	
	if (!cheap)
	{ // do the full Sobel operator
		float tl = unpack(tex2D(sSrc, UV + vTexOffsetsDiag[0].xy));
		float tr = unpack(tex2D(sSrc, UV + vTexOffsetsDiag[0].zw));
		float bl = unpack(tex2D(sSrc, UV + vTexOffsetsDiag[1].xy));
		float br = unpack(tex2D(sSrc, UV + vTexOffsetsDiag[1].zw));

		dX = -tl -2*l -bl + tr + 2*r + br; // approximate derivatives
		dY = -tl - 2*t -tr + bl + 2*b + br; // using Sobel operator
	}
	else
	{ // doesn't look as nice
		dX = -l + r;
		dY = -t + b;
	}
	
	// Compute cross-product and renormalize
	float3 N = normalize(float3(-dX, -dY, 1));
	// Convert signed values from -1..1 to 0..1 range and return
	return float4(N * 0.5 + 0.5, 1);
}

technique GenerateNormalMap
{
	pass p0
	{
		VertexShader = NULL;
		PixelShader = compile ps_2_0 ps_generate_normalmap(false);
	}
}
